#!/usr/bin/python

# tz_cms50.py
#       --copyright--                   Copyright 2011 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       September 19, 2011      bar
#       September 20, 2011      bar     figure out bottom 4 bits of 3rd byte (dupes of top 4 of 2nd byte, Y value)
#                                       8O1 instead of 8N1 - like PC program
#                                       fix timeouts
#                                       fix > 127 heart rates
#       September 21, 2011      bar     .png file output
#                                       put the sample getting inside the thing
#                                       spin off find_likely_COM_ports() to tz_usb.py
#       September 22, 2011      bar     continue to fight auto-upload and upload in general
#       September 23, 2011      bar     rememmber the a0, b0, c0 cmds that come in every 255th sample in upload
#                                       remember the hr==ox==0 samples
#                                       --verbose
#       September 24, 2011      bar     time stamp the file name of the main wave sample output file
#                                       run the .png file
#                                       auto-restart the com port in case it was yanked
#       September 25, 2011      bar     don't write .dat files and whack empty .plsoxi files after they've been written
#       September 27, 2011      bar     allow no io to be given at create time
#       September 28, 2011      bar     print average wave values or something else
#       September 29, 2011      bar     upload needs slower timeout
#       October 2, 2011         bar     put the USB/serial ids in constants at the top
#       October 4, 2011         bar     write_new_samples when flushing a recording file
#       October 10, 2011        bar     finger samples
#                                       rework sample parsing
#       October 11, 2011        bar     able to restart writing to a new file
#       October 14, 2011        bar     allow simple sample ox of 255 (as in recordings)
#       November 9, 2011        bar     be robust in the face of no pygooglechart installed
#                                       and allow graph with the finger "samples" in there
#       November 9, 2011        bar     try multiplying the low 3 bits of the 1st byte in to the waveform values
#       November 11, 2011       bar     only use one bit from byte 3 as the high bit of the heart rate - ignore the byte 3 high bit
#                                       except that the byte 3 high bit will cause us now to ignore the sample entirely
#       November 13, 2011       bar     try sending an E to the device (another device responds with version info)
#       November 29, 2011       bar     pyflake cleanup
#       May 27, 2012            bar     doxygen namespace
#       January 18, 2016        bar     except syntax change
#       December 16, 2016       bar     helpfull prompt
#       December 3, 2017        bar     put the 'when' in samples
#       January 25, 2018        bar     a doc string for a_finger_sample()
#       December 16, 2019       bar     parse samples with extra stuff at the ends of the sample text lines - to allow star-graph waveforms to be stored in the sample files
#                                       parse a float in the tm record at the top of a recording file
#       December 17, 2019       bar     sigh - would that I could code - and also, let's have yet another go at setting a recording time
#                                       write_graph option
#       December 19, 2019       bar     tweak tm at file close time - this whole logic needs a rethink. a_recording() is too smart and assumes too much
#       January 2, 2020         bar     print why the port didn't open
#       March 3, 2023           bar     python3
#                                       get rid of tz_google_chart stuff (it no longer works)
#       February 27, 2025       bar     the old device died, so now I have a CMS50E Hardware v 2.0.0 Software v 2.0.5 which has an entirely new protocol, etc. etc.
#       February 28, 2025       bar     put some code in for the E in cms50e device
#                                       more comments from the net
#       March 3, 2025           bar     --device_is_ok for testing
#       --eodstamps--
##      \file
#       \namespace              tzpython.tz_cms50
#
#
#       Communicate with a CMS50E Pulse Oximeter
#
#       Notes:
#
#           When recording, occasionally a USB stream has dropped samples - several. You can see the jump in the waveform.
#           When screen changes happen, ditto.
#           See the note: comments in upload code for information about what the device sends.
#
#
#   For CMS50E and beyond, put this in /etc/udev/rules.d after removing the #### from the ACTION we want:
#
    #   This file is named with a "z" so that the mode for the device is set by us rather than by pulseoxdl,
    #   which cuts "other" users out and sets the device's group plugdev, the latter of which is OK since
    #   we let everybody use the device.
    #
    #   Note to self:   Make sure any new file is a .rules file. Don't forget the extension.
    #                   And make sure the == and = situation isn't messed up anywhere in the rules.
    #                   And you gotta physically unplug this device to test this udev .rules file. Not just do this:
    #                       sudo udevadm control --reload
    #                       sudo udevadm trigger
    #                     # and:
    #                       /home/alex/bin/fix_keyboard.sh
    #
    #   And a note of where I found code to talk to this device:
    #       https://codeberg.org/klimd/pulseoxdl
    #
    #

    #   Why doesn't this work? It's not letter case.
    #   ACTION=="add", SUBSYSTEM=="hidraw", ATTR{idVendor}=="28E9", ATTR{idProduct}=="028A", MODE="0666", PROGRAM="/bin/sh -c 'echo $(($(/bin/ls -t1 /dev/hidraw_oxi_* 2>/dev/null | /bin/grep -om 1 [0-9]\\+$) + 1))'", SYMLINK+="hidraw_oxi_%c"
    ####ACTION=="add", SUBSYSTEM=="hidraw", DEVPATH=="*28E9:028A*",                          MODE="0666", PROGRAM="/bin/sh -c 'echo $(($(/bin/ls -t1 /dev/hidraw_oxi_* 2>/dev/null | /bin/grep -om 1 [0-9]\\+$) + 1))'", SYMLINK+="hidraw_oxi_%c"

    # eof

#
#   From https://ian.ahands.org/progs/pulseox/          With an expired LetsEncrypt thingee February 28, 2025
#
#       The CMS-50e sets up a USB serial character device. The character device speaks a very simple protocol.
#       There is basically a 5 byte stream/sequence of data:
#
#           (128-???)   The first byte has and interger value greater than 127, all other values
#                       seem to be 7-bit only (ignoring the most significant bit).
#           (0-127)     The next byte is the y-axis for the wave form.
#           (0-10 ?)    Unknown.
#           (0-127)     This byte is the detected heart rate. What is interesting here is that human
#                       hearts can easily beat > 127 bpm. So at 128 bpm the device sends a 0, but
#                       the device (and my program) adds 128 when displaying the number. There must
#                       be a cutoff somewhere (where the 128 stops being added), but I have been
#                       unable to detect where (I can't get my heart rate to 200, or <= 40. I have
#                       assumed the cutoff is 40...
#           (0-99)      This byte is the detected oxygen saturation.
#
#       BAR note: other places describe the HR byte and how they allow higher rates than 127.
#
#

#
#
#       TODO:
#           Jeez. Why didn't I put a full float time stamp at the top of the file and then just use elapsed times subsequently? Yeah. Reading the file couldn't be done by just randomly jumping in to the file and starting after the next \n, but ... that's easy to fix.
#
#           figure out what the 6 low bits in the 1st byte are
#               0x08 and 0x10 seem to be triggered by valsalva - PC program says "Searching" and flat-lines the waveform display
#               it's the low nibble values of 5, 6, and 7 that come through for no easy-to-understand reason
#           figure out the a0, b0, c0 cmds in upload data
#           figure out what the top 0x30 bits in the 3rd byte are (top 3 bits have been detected set, but the top bit is not necessarily what we make of it (hr + 256))
#           figure out commands to device (use \winbin\portmon on PC - note: cannot run from a net drive)
#               Any byte sent to it seems to turn on USB streaming (f5 and f6 do it - some others did, too)
#
#           Upload from portmon:
#               >> f5 f5            These seem to also be sent out in the middle of the dump from the device
#               << 86 09 01
#               << 47 5f
#               << 86 08 01
#               << 47 5f
#               .
#               .
#               .
#               >> f6 f6 f6
#
#

from    __future__  import  print_function

import  fcntl
import  glob
import  math
import  os
import  re
import  socket
import  sys
import  time

import  serial              # from pySerial

import  output_files
import  tz_usb
import  tzlib


USB_VENDOR_ID       = 0x10c4
USB_PRODUCT_ID      = 0xea60
USB_VENDOR_ID_EF    = 0x28E9                # for CMS50E or F
USB_PRODUCT_ID_EF   = 0x028A


SAMPLE_RATE     = 60


class   a_cms50_exception(Exception) :
        pass

class   a_cms50_timeout_exception(a_cms50_exception) :
        pass                                # there's nothing to be had for now

class   a_cms50_data_exception(a_cms50_exception) :
        pass                                # the data isn't ready yet

class   a_cms50_no_finger_exception(a_cms50_exception) :
        pass                                # a "finger" isn't in the device


def find_all_cms50e_devices() :
    return(glob.glob('/dev/hidraw_oxi_*') or glob.glob('/dev/cms50e*'))


def print_device_list(port_list) :
    print(find_all_cms50e_devices() or tz_usb.find_likely_COM_ports(vendor_id = USB_VENDOR_ID, product_id = USB_PRODUCT_ID, list_level = port_list))


def find_first_device(noisy = False) :
    """ Find the first of any CMS50 devices we might have. Return None if none found. """
    port    = None
    nports  =                       tz_usb.find_likely_COM_ports(vendor_id = USB_VENDOR_ID, product_id = USB_PRODUCT_ID,  serial_number = "ca8d")     # kludge to avoid cardiox driver board in one incarnation
    ports   =                       find_all_cms50e_devices()
    ports   = ports or [ p for p in tz_usb.find_likely_COM_ports(vendor_id = USB_VENDOR_ID, product_id = USB_PRODUCT_ID ) if p not in nports ]
    if  len(ports) :
        port    = ports[0]
        if  noisy :
            print(  "Using port", port,)
        if  (len(ports) > 1) and noisy :
            print("Found ports", ports, end = ' ')
        if noisy :
            print('')
        pass
    return(port)



class   a_simple_sample(object) :

    p_attrs = [ 'hr', 'ox', ]

    def __init__(me, hr = 0, ox = 0) :
        me.hr   = hr
        me.ox   = ox
        if  (me.ox > 255) :                             # in practice, notice the 'when' value in a_finger_sample() is gonna always be over 255 (and a dot-float, to boot, but we don't even need to check that) - anyway, we'll not want the 'when' to be 0, though the code allows it
            raise ValueError("ox:%s" % str(me.ox))      # also, 'ox' value is sometimes 255 and there's a bogus 'hr' value going with it - possibly 127, for instance - also, it appears [0,0] samples are placeholders while the finger is out
        if  me.hr is None :
            raise ValueError("hr:%s" % str(me.hr))
        pass

    def print_str(me) :
        return("%2u%% %3ubpm" % ( me.ox, me.hr ) )

    def __str__(me) :
        aa      = [ getattr(me, atr, None) for atr in me.p_attrs ]
        return(repr(aa))

    #   a_simple_sample



class   a_finger_sample(a_simple_sample) :
    """ A "sample" that indicates when a finger goes on or off. """

    p_attrs = [ 'on', 'when', ]

    def __init__(me, on = False, when = None) :
        me.when     = when or time.time()
        me.on       = on

    def print_str(me) :
        return("Finger %s %f %s" % ( ((me.on and "on") or "off"), me.when, time.strftime('"%H:%M:%S %b %d, %Y"', time.localtime(me.when)), ) )

    #   a_finger_sample



class   a_full_sample(a_simple_sample) :

    csv_header  = '"Sample", "Beat", "Value", "Oxi%", "BPM", "?b1", "?b3hi"'

    p_attrs     = a_simple_sample.p_attrs + [ 'num', 'y', 'beat', 'bc', 'ac', ]

    def __init__(me, hr = 0, ox = 0, num = 0, y = 0, beat = False, bc = 0, ac = 0) :
        super(a_full_sample, me).__init__(hr = hr, ox = ox)
        me.num  = num
        me.y    = y
        me.beat = beat

        me.bc   = bc            # low 6 bits of 1st byte
        me.ac   = ac            # 0x30 bits of 3rd byte

        # me.y    = ((me.y) * ((me.bc & 0xf) + 1)) / 4.5

    def csv_str(me) :
        return("%d, %s, %d, %d, %d, %d, %d"             % ( me.num, ((me.beat and '"B"') or ''),  me.y, me.ox, me.hr, me.bc, me.ac, ) )

    def print_str(me) :
        return("%u: b=%02x ac=%02x %s y=%3d %s"    % ( me.num, me.bc, me.ac, ((me.beat and 'B') or ' '), me.y, super(a_full_sample, me).print_str(), ) )

    #   a_full_sample


def _parse_sample(s) :
    try :
        return(eval(s))
    except ( TypeError, ValueError ) :
        pass
    return(None)

def _create_sample(aa, cls) :
    if  len(aa) == len(cls.p_attrs) :
        kwargs  = dict(zip(cls.p_attrs, aa))
        try :
            return(cls(**kwargs))
        except ( TypeError, ValueError ) :
            pass
        pass
    return(None)


def parse_sample(s) :
    aa  = _parse_sample(s)
    ss  = _create_sample(aa, a_full_sample) or _create_sample(aa, a_simple_sample) or _create_sample(aa, a_finger_sample)
    return(ss)


class   a_recording(object) :

    FILE_EXT            = ".plsoxi"

    def __init__(me, tm = None, samples = [], fd = "", write_graph = False) :
        if  tm is None  :
            tm          = time.time()
        me.tm           = tm
        me.now          = tm                                        # the time of the next sample we'll get
        if  me.now      :
            scnt        = 0.0
            tm          = sys.maxsize
            for s in samples :
                when    = getattr(s, 'when', None)
                if  when is None :
                    s.when  = me.now + (scnt / SAMPLE_RATE)
                else        :
                    me.now  = s.when
                    scnt    = 0.0
                if  hasattr(s, 'y') :
                    scnt   += 1.0
                    tm      = min(tm, s.when)                       # find the earliest sample time
                pass
            if  tm < sys.maxsize :
                me.tm   = tm                                        # make our time the earliest real sample time
            me.now      = me.now + (scnt / SAMPLE_RATE)             # set the time it'll be when we get the next sample
        me.samples      = samples or []                             # a_???_sample()s
        me.fd           = fd      or ""                             # raw data
        me.fo           = None                                      # file we are outputting to, if any
        me.write_graph  = write_graph                               # whether we put out the waveform stars in the file

    def forget_old_samples(me, how_many) :
        how_many        = max(0, int(how_many))
        me.samples      = me.samples[how_many:]                     # let's not run out of memory if this thing runs for days
        if  me.fo       :
            me.wfsi     = max(0, me.wfsi - how_many)
        pass

    def append(me, sample) :
        sample.when     = getattr(sample, 'when', me.now)
        me.now          = sample.when                               # get back in sync so we don't drift
        if  not hasattr(sample, 'when') :                           # note: we don't do this logic now - why? Because we're just stepping me.now and we've set it at the start
            sample.when = time.time()                               # rather than using 'now', for now, just assume we're appending real-time samples. yuck.
        if  hasattr(sample, 'y') :
            me.tm       = min(me.tm, sample.when)                   # set our .tm to the earliest real sample time we've seen
            me.now     += 1.0 / SAMPLE_RATE                         # set the time it'll be when we get the next sample
        me.samples.append(sample)

    def tm_str(me) :
        if  me.tm < 3600 * 24 * 365 :
            tm  = me.tm % ( 3600 * 24 )
            return("%02u:%02u" % ( tm // 3600, (tm // 60) % 60 ) )  # hour:minute
        return(time.strftime('"%H:%M:%S %b %d, %Y"', time.localtime(me.tm)))


    def flush_file(me) :
        me.write_new_samples()
        if  me.fo :
            me.fo.flush()
        pass

    def write_new_samples(me) :
        if  not me.fo :
            return(False)

        nsc         = len(me.samples)
        cnt         = nsc   - me.wfsi
        for sa in me.samples[me.wfsi:nsc] :
            if  me.wfna    != sa.p_attrs :
                me.wfna     = sa.p_attrs
                me.fo.write("fields %s\n" % str(me.wfna))
            if  me.write_graph and hasattr(sa, 'y') :
                ys  = (' ' * int((100.0 * (sa.y - 0)) / max(1.0, (128 - 0)))) + '*'
                me.fo.write("s %-35s | %s\n" % ( str(sa), ys, ))
            else    :
                me.fo.write("s %s\n" % str(sa))
            pass

        me.wfsc    += cnt
        me.wfsi     = nsc
        return(cnt)

    def close_write_file(me) :
        if  me.fo   :
            me.write_new_samples()
            me.fo.close()
            me.fo   = None
            if  not me.wfsc :
                os.remove(me.fn)                        # whack the file if there were no samples written
            if  len(me.samples) :
                me.tm   = max(me.tm, me.samples[-1].when)
            else        :
                me.tm   = max(me.tm, time.time())
            return(True)
        return(False)

    def open_write_file(me, fn, only_new_samples = False) :
        me.close_write_file()

        me.wfsi     = (only_new_samples and len(me.samples)) or 0
        me.wfsc     = 0
        me.wfna     = []
        me.fn       = os.path.splitext(fn)[0] + me.FILE_EXT         # force the proper ext (?)

        me.fo       = output_files.a_file(me.fn)
        me.fo.write("tm %f %s\n\n" % ( me.tm, me.tm_str() ) )
        return(me.fn)


    def write_file(me, fn) :
        fn      = me.open_write_file(fn)
        if  me.close_write_file() :
            return(fn)
        return(None)


    s_vals_parse_re = re.compile(r'^s\s+(\[.+\])', re.DOTALL)

    @staticmethod
    def parse(fd, write_graph = None) :
        samples = []
        for li in re.split(r"\r?\n", fd) :
            g   = a_recording.s_vals_parse_re.search(li)
            if  g   :
                s   = parse_sample(g.group(1))
                if  s :
                    samples.append(s)
                pass
            pass
        if  len(samples) :
            g   = re.search(r"\ntm (" + tzlib.float_regx_str + r")", fd)
            if  g :
                tm  = float(g.group(1))
                return(a_recording(tm, samples, write_graph = write_graph))
            pass

        return(None)


    @staticmethod
    def parse_file(fn, write_graph = None) :
        fd  = tzlib.read_whole_text_file(fn)
        return(a_recording.parse(fd, write_graph = write_graph))

    #   a_recording



def is_serial_device(dev) :
    """ Is this object a Serial port? """
    return(str(dev.__class__).find("Serial") >= 0)


class   a_comm(object)  :

    def __init__(me, io = None, timeout = 0.01) :
        me.io           = io
        me.timeout      = timeout
        me.ibuf         = ""
        me.scnt         = 1
        me.rx_when      = tzlib.elapsed_time()  - 10000.0
        me.data         = []                    # queue of a_recording's driven by data coming in from the device
        me.ping_when    = 0
        me.tx_when      = 0                     # when we most recently transmitted to the device (we throttle our outout)
        me.hr           = 0                     # most recent heart rate
        me.ox           = 0                     # most recent oxi value

        if  not me.io   :
            me.reopen()
        pass


    def is_serial_device(me) :
        return(is_serial_device(me.io))


    def flush(me) :
        io      = me.io
        if  io  :
            try :
                io.flush()
            except ( socket.error, OSError, IOError, AttributeError, ) :
                raise a_cms50_exception("Flush error!")
            pass
        pass


    def tx(me, s) :
        td  = tzlib.elapsed_time() - me.tx_when
        if  td < 0.1 :
            time.sleep(td)
            me.tx_when  = tzlib.elapsed_time()
        s   = tzlib.convert_to_latin1_bytes(s)
        # print("@@@@        _tx", len(s), [ hex(tzlib.safe_ord(cc)) for cc in s ], s)
        return(me.io.write(s))


    def rx(me, how_many = 1) :
        try :
            return(me.io.read(how_many).decode('latin1'))
        except ( serial.SerialException, serial.serialutil.SerialException, OSError, IOError, AttributeError, ) :
            pass
        return('')


    def empty_rx(me) :
        me.ibuf = ""
        tt      = tzlib.elapsed_time()
        while True :
            t   = tzlib.elapsed_time()
            r   = me.rx(1)
            if  len(r) and False :
                tt  = t
            elif t - tt > 0.1 :
                break
            pass
        pass


    def rx_until(me, how_many = 1, timeout = 0) :
        """ Wrote this and then saw the read() rtn just below. """
        timeout = timeout or 0.1
        t       = tzlib.elapsed_time()
        buf     = b''
        while how_many  and (tzlib.elapsed_time() - t >= timeout):
            r   = me.rx(how_many)
            if  len(r)  :
                t       = tzlib.elapsed_time()
            how_many   -= len(r)
            buf        += r
        return(buf)


    def read(me, how_many = 1, timeout = None) :
        timeout         = timeout or me.timeout

        t               = nt    = tzlib.elapsed_time()
        r               = ""
        while True      :
            rr          = me.ibuf[:how_many]                # pull in anything we have buffered - what we've read from the device, but haven't returned from this routine
            me.ibuf     = me.ibuf[how_many:]
            how_many   -= len(rr)
            r          += rr
            if  not how_many :
                break

            rr          = me.rx(how_many)
            me.ibuf    += rr
            if  len(rr) :
                t       = nt                                # don't time out if something is coming in

            if  nt - t >= timeout :
                if  len(r) :
                    break                                   # give the caller what we got before timing out
                print("@@@@ no read", tzlib.elapsed_time(), nt, t, nt - t, timeout)
                raise a_cms50_timeout_exception("read")

            nt          = tzlib.elapsed_time()

            if  not len(rr) :                               # don't hammer the device if we're not getting anything
                time.sleep(min(0.1, timeout / 2.0))
            pass

        # print("rxing", len(r), "%04x" % ( len(r) ), hexify(r))

        return(r)



    def read_upload(me, progress_rtn = None, verbose = 0) :
        timeout     = 10
        while True  :
            try     :
                r   = me.read(1, timeout = timeout)
                timeout = 0.1
                a1  = me.read(1, timeout = timeout)
                a2  = me.read(1, timeout = timeout)
                cmd = ord(r)
                if  (cmd == 0xf2) and (ord(a1) & 0x80) :
                    me.ibuf = r + a1 + a2 + me.ibuf
                    break
                elif cmd == 0xf0 :
                    raise a_cms50_data_exception("No upload header")
                else :
                    me.ibuf = a1 + a2 + me.ibuf
            except ( a_cms50_exception, a_cms50_timeout_exception, ) :
                return(None)
            pass

        fd          = ""
        tm          = 0
        prv         = 0
        mxc         = 10000000
        samples     = []
        while True  :
            try     :
                r   = me.read(1, timeout = timeout) + me.read(1, timeout = timeout) + me.read(1, timeout = timeout)
                cmd = ord(r[0])
                hr  = ord(r[1])
                ox  = ord(r[2])
                if  not (cmd & 0x80) :
                    # print("@@@@ bad cmd %02x:%02x:%02x" % ( cmd, ord(r[1]), ord(r[2]) ))
                    break

                sa  = None

                if  (not hr) and (not ox) :
                    # a = hr | (ox << 8)
                    # print("@@@@ @%u toss out msg: 2nd/3rd no hi bit cmd %02x arg=%02x:%02x d=%d x=%04x" % ( len(fd) // 3, cmd, hr, ox, a, a ))
                    sa  = a_full_sample(hr, ox, bc = cmd)
                    r   = ""                                        # this may be a bug in the device. not sure. not sure whether tis correct to toss the data out of mxc tracking. Too, a ox of 0xff came up in a recording that had 6 extra samples even after this, which happened at 255-bytes in.
                elif cmd == 0xf2 :
                    tm  = 60 * ((60 * (hr & 0x1f)) + ox)
                    prv = 0xf2
                else    :
                    if  prv == 0xf2 :
                        mxc = ((cmd & 0x3f) << 14) | ((hr & 0x7f) << 7) | ox
                        # print("@@@@ @%u mxc=%u %02x:%02x:%02x" % ( len(fd), mxc, cmd, hr, ox, ))
                        mxc = mxc + len(fd) - 9
                    elif ((cmd & 0xf0) != 0xf0) :
                        # a         = hr | (ox << 8)
                        # print("@@@@ info@%u: cmd %02x arg=%02x:%02x le:%d:0x%04x" % ( len(fd), cmd, hr, ox, a, a ))
                        if  (cmd   == 0x80) and (len(fd) + 3 >= mxc) :
                            me.ibuf = r + me.ibuf
                            r       = ""                                # !!!! kludge to remove ending, regular, 5-byte samples
                            mxc    -= 3
                        elif (cmd  == 0x80) and (hr == 0) :
                            sa      = a_full_sample(hr, ox, bc = cmd)
                            mxc    -= 3                                 # !!!! is this sample a reflection of a glitch in the recording? and, if so, why doesn't the byte count reflect it? (or does it mark something? or what?)
                        else        :
                            sa      = a_full_sample(hr, ox, bc = cmd)   # note: every 256th sample (after the 1st, 253rd - 256, including the two F2 and the 8x length "samples") is one of these - cmds are in [ 0xa0, 0xb0, 0xc0, ], hr and oxi are normal
                        pass
                    else    :
                        hr          = ((cmd & 0x3) << 7) | (hr & 0x7f)
                        sa          = a_simple_sample(hr, ox)           # note: in each 256 sample batch, the 85th sample and the 170th sample are ox==255
                    prv = cmd

                if  sa  :
                    samples.append(sa)

                fd += r
                if  progress_rtn :
                    progress_rtn(len(fd), mxc)

                if  len(fd) >= mxc :
                    # print("@@@@ ctmxc", len(fd), mxc)
                    break
                pass
            except ( a_cms50_exception, a_cms50_timeout_exception, ) :
                # print("@@@@ timeout")
                break
            pass

        if  progress_rtn :
            progress_rtn(len(fd), len(fd))

        # print("@@@@", len(fd), "of", mxc, "   ", len(fd) // 3, "msgs of", (mxc + 2) // 3)

        return(a_recording(tm = tm, samples = samples, fd = fd))



    def _read_serial_sample(me, progress_rtn = None, verbose = 0, mismatch_callback = None) :
        s       = None
        try     :
            b           = ord(me.read())
            me.rx_when  = tzlib.elapsed_time()

            if  b  == 128 :
                ba  = [ b ]
                t   = me.rx_when
                while (len(ba) < 5) and (t - me.rx_when < 0.1) :
                    t   =   tzlib.elapsed_time()
                    ba.append(ord(me.read()))

                me.scnt    += 1

                if  False and (ba[1] in [ 0, 0xff ]) :              # 321390_Kanograf_(8400).pdf
                    if  len(ba) == 5 :
                        print("@@@@ %02x:%02x:%02x:%02x:%02x" % ( ba[0], ba[1], ba[2], ba[3], ba[4] ))
                    else    :
                        print("@@@@", str(ba))
                    pass

                raise a_cms50_no_finger_exception

            if  b   < 128 :
                # print("@@@@ toss read %02x" % b)
                pass
            else    :
                y   = ord(me.read())

                if  (b == 0xf2) and (y & 0x80) :                    # read_upload situation?
                    me.ibuf = chr(b) + chr(y) + me.ibuf             # unread the two bytes we have
                    dt      = me.read_upload(progress_rtn = progress_rtn, verbose = verbose)
                    if  dt  :
                        me.data.append(dt)

                    return(None)

                ay  = ord(me.read())
                hr  = ord(me.read())
                if  (ay == 0xf2) and (hr & 0x80) :                  # after dropping two bytes (in b and y), is this a read_upload situation?
                    me.ibuf = chr(ay) + chr(hr) + me.ibuf           # unread the second two bytes
                    dt      = me.read_upload(progress_rtn = progress_rtn, verbose = verbose)
                    if  dt  :
                        me.data.append(dt)

                    return(None)

                hr |= ((ay & 0x40) << 1)

                ox  = ord(me.read())

                if  ay & 0x80 :
                    # me.ibuf = chr(ay) + chr(hr) + chr(ox) + me.ibuf       # note: the few times this has happened have not indicated that this is a good idea - nor that interpreting the ay|hr|ox bytes as uploaded data is a good idea
                    return(None)                                            # punt as best we can

                bc  = (b & ~(64 | 128))                                     # remember the low 6 bits of the first, command byte
                ac  = ay >> 4                                               # drop the 4 bit on-screen bar value?
                # print("@@@@ %02x" % b, hr, ox, me.scnt, y, ay, bc, ac)
                s   = a_full_sample(hr, ox, me.scnt, y, ((b & 64) and True) or False, bc = bc, ac = ac & 3)

                if  (ay & 0x0f) != (y // 8) :           # apparently, these 4 bits are a dupe of the top 4 or the 7 bits in "y" - let's force that to be so - crashes when device dumps memory
                    # print("@@@@ b=%02x ay & 0xf != y/8  ay:0x%02x != y:0x%02x" % ( b, ay, y )           # can happen if upload is starting)
                    if  mismatch_callback :
                        mismatch_callback(s, b, ay)     # tell the caller this happened if he wants to know
                    s       = None

                me.scnt    += 1
            pass

        except ( a_cms50_exception, a_cms50_timeout_exception, ) :
            if  tzlib.elapsed_time() - me.rx_when > 0.5 :
                me.scnt     = 1
            raise

        return(s)


    def _read_dev_sample(me, progress_rtn = None, verbose = 0, mismatch_callback = None) :
        s           = None
        try         :
            t       = tzlib.elapsed_time()
            if  t - me.ping_when > 4.5 :
                me.ping_when    = t
                me.send_cmd([ 0x82, 0x02, ])    # ping
                me.flush()

            # print(".", end = '')
            buf         = me.read(1, timeout = 1.1)         # yeah, the device goes away every 30 seconds and there doesn't seem to be anything I can do about it. maybe some faster pinging would help ???? !!!!
            me.rx_when  = tzlib.elapsed_time()
            cmd         = ord(buf[0])
            if  cmd == 0xeb :
                buf    += me.read(5)
                c       = ord(buf[1])
                if  not c :                     # amplitudes msg? 60 per second?
                    if  me.checksum(buf[:-1]) != ord(buf[-1]) :
                        print("@@@@ Bad eb:00 checksum", [ ord(cc) for cc in buf ])
                        return(None)
                    # print("@@@@ y", [ ord(cc) for cc in buf ])
                    if  ord(buf[4]) ==  0x60 :
                        raise a_cms50_no_finger_exception
                    if  me.hr and me.ox :
                        y               = ord(buf[3])   # !!!! (ac buf[2]) is the multiplier index along with the BEAT detector)
                        beat_detected   = ((ord(buf[2]) & 0x40) != 0)
                        s               = a_full_sample(me.hr, me.ox, me.scnt, y, beat_detected, bc = ord(buf[4]), ac = ord(buf[2]))        # note: bc is the 0..15 bar-chart thingee
                        me.scnt        += 1

                        return(s)

                    else :
                        # me.ask_for_dev_hroxy()    # the device does not like this
                        pass

                    pass

                elif c == 0x01      :               # measurements msg? once a second? also when change????
                    buf            += me.read(2)
                    if  me.checksum(buf[:-1]) != ord(buf[-1]) :
                        print("@@@@ Bad eb:01 checksum", [ ord(cc) for cc in buf ])
                        return(None)
                    # print("@@@@ hrox", [ ord(cc) for cc in buf ])
                    if  ord(buf[4]) == 0x7f :
                        raise a_cms50_no_finger_exception               # this is the normal way we detect finger-out
                    me.hr           = ((ord(buf[2]) & 0x2) and (ord(buf[3]) + 0x80)) or ord(buf[3])
                    me.ox           = ord(buf[4])
                    print("@@@@ hr|ox", me.hr, me.ox, [ ord(cc) for cc in buf ])

                elif c == 0x7f :                    # stopped streaming
                    print("@@@@ dev stopped streaming")
                    me.ask_for_dev_hroxy()

                else :
                    print("@@@@ unknown 0xeb and:", c, hex(c))
                    raise a_cms50_no_finger_exception

                # print("\rSpO2: %3d%%, PR: %3d BPM, PB: |%-16s|" % ( spo2, pr, pulsebars[pulsebar & 0xf], ))
            elif cmd   == 0xf2 :                # reply to our ping?
                buf    += me.read(7)
                if  me.checksum(buf[:-1]) != ord(buf[-1]) :
                    print("@@@@ Bad ping/0xf2 checksum", [ ord(c) for c in buf ])
                    return(None)
                else    :
                    # print("@@@@ ping reply", [ ord(c) for c in buf ])
                    pass
                # me.ask_for_dev_hroxy()
                pass
            elif cmd    :                       # ignore the 64-byte-packet filler nulls
                print("@@@@ cmd:", cmd, hex(cmd))
                pass
            pass
        except a_cms50_no_finger_exception :
            # print("@@@@@@@@@ nf")
            me.hr   = 0
            me.ox   = 0
            raise                               # tell the caller the finger is not in the device
        except a_cms50_timeout_exception :
            # print("@@@@@@@@@ timeout except")
            if  tzlib.elapsed_time() - me.rx_when > 0.5 :
                me.scnt     = 1
            raise
        except a_cms50_exception :
            # print("@@@@@@@@@ generic exception")
            if  tzlib.elapsed_time() - me.rx_when > 0.5 :
                me.scnt     = 1
            raise

        return(s)


    def read_sample(me, progress_rtn = None, verbose = 0, mismatch_callback = None) :
        if  not me.io : return(None)

        if  me.is_serial_device() :
            return(me._read_serial_sample(progress_rtn = progress_rtn, verbose = verbose, mismatch_callback = mismatch_callback))
        return(    me._read_dev_sample(   progress_rtn = progress_rtn, verbose = verbose, mismatch_callback = mismatch_callback))


    def close(me)   :
        if  me.io   :
            c       = me.io
            me.io   = None
            try     :
                c.close()
            except ( socket.error, OSError, IOError, AttributeError, ) :
                raise a_cms50_exception("Close error!")
            pass
        pass



    def blind_close(me) :
        try :
            me.close()
        except a_cms50_exception :
            pass
        pass


    def reopen(me, port = None) :
        me.blind_close()

        me.hr   = 0
        me.ox   = 0

        port    = port or find_first_device(noisy = True)
        port    = port or '/dev/hidraw_oxi_1'                           # this device doesn't show up unless the finger is in and it's reading ok - so we'll default to it in the blind

        if  not port :
            return(None)

        try :
            port    = int(port)
            cport   = port - 1
        except ValueError :
            cport   = port

        try :
            me.io   = serial.Serial(port = cport, baudrate = 19200, parity = "O", timeout = 0.001)                # note: PC program sets 8O1. serial.Serial() multiplies timeout by 1000 before passing to windows (This 1 mill is minimum for windows. I don't know about other OS's.)
        except serial.SerialException :
            try     :
                me.io       = open(cport, "r+b")
                oldflags    = fcntl.fcntl(me.io, fcntl.F_GETFL)
                fcntl.fcntl(me.io, fcntl.F_SETFL, oldflags | os.O_NONBLOCK)
                me.timeout  = max(me.timeout, 0.200)
            except  ( OSError, IOError, AttributeError, ) :
                return(False)
            pass

        return(True)


    @staticmethod
    def checksum(s) :
        return(sum([ ord(c) if tzlib.is_stringish(c) else c for c in s ]) & 0x7f)


    def send_cmd(me, cmd) :
        # print("@@@@ sendcmd_in", len(cmd), [ hex(cc) for cc in cmd ], cmd)
        cs  = cmd[-1]
        if  cs < 0x80   :                                           # the stop-sending-data string is a bunch of 0x80s plus some other things. That cmd doesn't have a checksum at the end. Just another 0x80
            if  cs != me.checksum(cmd[:-1]) :
                print("@@@@ our bad cs", "his", cs, "ours", me.checksum(cmd[:-1]), "cmd", [ "%3d 0x%2x" % ( c, c, ) for c in cmd ])
                return(None)
            pass
        d   = '\x00' + "".join([ chr(c) for c in cmd ]) + ('\x00' * (64 - len(cmd)))
        me.tx(d)
        me.flush()
        time.sleep(0.1)
        return(d)


    def send_cmd_and_rx(me, cmd, rx_len) :      # this command has the checksum already in  ... ... maybe a zero before the 64 bytes might get the device out of a weird state. Doesn't hurt. But normally not needed
        d   = me.send_cmd(cmd)
        d                                       # for pyflakes
        r   = me.read(64, timeout = 0.5)[:rx_len]
        # print("@@@@", len(r), [ hex(ord(c)) for c in r ], hex(ord(r[rx_len - 1])), ord(r[rx_len - 1]), me.checksum(r[:rx_len - 1]), len(d), [ hex(ord(c)) for c in r[:rx_len - 1] ], )
        if  me.checksum(r[:rx_len - 1]) != ord(r[rx_len - 1]) :
            print("@@@@ bad checksum. ours", me.checksum(r[:rx_len - 1]), "his", ord(r[rx_len - 1]))
            r   = None      # !!!!
        return(r)


    def ask_for_dev_hroxy(me) :
        if  me.io and (not me.is_serial_device()) :
            me.send_cmd([ 0x9b, 0x01, 0x1c, ])      # ask for streaming measurements  hr and ox
            me.flush()
            me.send_cmd([ 0x9b, 0x00, 0x1b, ])      # ask for streaming amplitudes    y
            me.flush()
        pass

    def stop_dev_hroxy(me) :
        if  me.io and (not me.is_serial_device()) :
            me.send_cmd([ 0x9b, 0x7f, 0x1a, ])
            me.flush()
        pass


    def fire_up_device(me) :
        """ Send the initial stuff to the device. """
        if  me.io :
            # print("@@@@ firing up device");

            me.empty_rx()
            try :
                me.send_cmd_and_rx([ 0x7d, 0x81, 0xa7, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x7d, 0x81, 0xa2, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,  ],  2)  # stop sending data (no checksums here, and what's with the 7d?)
            except  a_cms50_timeout_exception as e :
                if  e.message == 'read' :
                    me.stop_dev_hroxy()
                pass
            me.empty_rx()
            me.send_cmd_and_rx([ 0x82, 0x02,      ],  8)    # ping
            me.send_cmd_and_rx([ 0x80, 0x00,      ],  2)    # ???
            me.send_cmd_and_rx([ 0x8e, 0x03, 0x11 ], 10)    # user name
            me.send_cmd_and_rx([ 0x81, 0x01,      ], 10)    # model name

            tm  = time.localtime(time.time())
            cmd = [ 0x83, (tm.tm_year - 2000) & 0x7f, tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, 0x00, 0x00 ]
            cmd.append(me.checksum(cmd))
            me.send_cmd_and_rx(cmd, 3)                      # set the device's date/time clock

            time.sleep(.1)
            me.ask_for_dev_hroxy()

            # print("@@@@ device fired up");
        pass


    def start_usb(me) :
        if  me.io   :
            try     :
                if  is_serial_device(me.io) :
                    me.tx("\xf5")       # try to get the USB streaming going again in case it's off
                else :
                    me.fire_up_device()
                pass
            except ( serial.SerialException, OSError, IOError, AttributeError, ) :
                tzlib.print_exception()
                me.reopen()
            pass
        pass

    pass        # a_comm



def _program_version_str(ident) :
    return("V" + re.sub(r".*<program_version>([\d\.]+)</program_version>.*", r"\1", ident))



def get_output_file_name(ofile_name, program_name = None, ext = a_recording.FILE_EXT) :
    if  not ofile_name :
        ofile_name  = os.path.basename(program_name or __file__)
    ofile_name      = os.path.splitext(ofile_name)[0] + ("_%010u%s" % ( int(time.time()), ext ))

    return(ofile_name)


def hms(t) :
    it  = int(t)
    return("%2u:%02u:%04.1f" % ( it // 3600, (it // 60) % 60, t - ((it // 60) * 60) ) )




class   a_progress_rtn(object) :
    def __init__(me) :
        me.pc   = -1
        me.sh   = False

    def show_progress(me, how_many, to_do) :
        pc  = (100  * how_many) // max(1, to_do)
        if  me.pc  != pc :
            me.pc   = pc
            if  how_many != to_do :
                me.sh   = True
                sys.stdout.write("%3u%%\r" % me.pc)
                sys.stdout.flush()
            elif me.sh :
                sys.stdout.write('%3u%%\n' % me.pc)
                sys.stdout.flush()
            pass
        pass

    #   a_progress_rtn


def main(ident) :
    import      TZCommandLineAtFile
    import      TZKeyReady
    import      tz_timer

    if  ident :
        sys.argv.insert(1, ident)
        sys.argv.insert(1, '--ident')

    program_name    = sys.argv.pop(0)

    TZCommandLineAtFile.expand_at_sign_command_line_files(sys.argv)


    ident       = ""
    port        = 0             # my COM port, not yours
    port_list   = 0
    verbose     = 0
    how_long    = float(sys.maxsize)
    device_ok   = False


    help_str    = """
%s (options) (output_files_base_name)

    I get the streaming data from a CMS50E Pulse Oximeter.

Options:

    --port  port_number     Set the COM port number
    --port_list             List possible ports (twice, list all available ports)
    --version               Print the program version number.
    --how_long  seconds     Only run for this long.            (default: forever)
    --device_is_ok          Run even if the device isn't available


""" % ( os.path.basename(program_name) )


    oi  = tzlib.array_find(sys.argv, [ "--help", "-?", "?", "-h", "/h", "/?", "?" ] )
    if  oi >= 0 :
        print(help_str)
        sys.exit(254)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--ident", "-i" ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        if  (oi >= len(sys.argv)) or not len(sys.argv[oi]) :
            print("Program IDENT info not given!")
            sys.exit(101)
        ident       = sys.argv.pop(oi)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--version", "-v" ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        print("version", _program_version_str(ident))
        if  not sys.argv :
            sys.exit(0)
        pass


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--port", "-p" ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        if  (oi >= len(sys.argv)) or not len(sys.argv[oi]) :
            print("No COM port given!")
            sys.exit(102)
        port        = sys.argv.pop(oi)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--port_list", "--pl", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        port_list  += 1


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--device_is_ok", "--device_ok", "--ok", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        device_ok   = True


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--how_long", "--how-long", "--howlong", ] )
        if  oi < 0  : break
        del sys.argv[oi]
        how_long    = float(sys.argv.pop(oi))

    while True :
        oi  = tzlib.array_find(sys.argv, [ "--verbose" ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        verbose    += 1



    ofile_name      = None
    if  len(sys.argv) >= 1 :
        ofile_name  = sys.argv.pop(0)

        if  ofile_name.startswith('-') :
            print("Put the whole path or a dot/slash before the output file name. Dashes are confusing: [%s]" % ofile_name)
            sys.exit(104)

        pass

    if  len(sys.argv) :
        print("I don't understand %s (--help for options)!" % ( sys.argv ))
        sys.exit(104)


    if  port_list   :
        print_device_list(port_list)


    me  = a_comm()
    if  (not me.io) and (not device_ok) :
        print("Please tell me a COM port to use with the --port option ( e.g. '--port 2' or '--port /dev/hidraw_oxi_1' )!")
        if  port :
            print("Port %s [if a number, then one lower port number] cannot be opened!" % port)
            print("A guess: Make this user able to open the serial port with: 'sudo usermod -a -G dialout ${USER}'")
        sys.exit(103)


    samples = a_recording()

    if  ofile_name :
        if  os.path.splitext(ofile_name)[1].lower() == ".csv" :
            fo  = output_files.a_file(ofile_name)
            fo.write(a_full_sample.csv_header + "\n")
        else    :
            fo  = None
            fn  = get_output_file_name(ofile_name, program_name = program_name)
            print("Outputting to: ", samples.open_write_file(fn))
        pass

    prg     = a_progress_rtn()
    stopped = 1000
    finger  = False
    lay     = []
    mx      = -10000
    mn      =  10000
    msa     = [ 64 ] * 300
    mss     = float(sum(msa))
    ts      = tzlib.elapsed_time()
    rx_when = ts
    png_drs = [ 7.5, 15.0, 30.0, 60.0, 120.0, 10 * 60.0, 15 * 60.0, 60 * 60.0, 2 * 60 * 60.0, 6 * 60 * 60.0, 12 * 60 * 60.0, ]
    png_di  = 0
    sb      = 0xf5

    try     :
        me.start_usb()              # in case he's not turned it on (though we'll do this every half second of silence from the device, anyway
    except a_cms50_timeout_exception :
        pass

    print("Type ? for help")

    start_time  =   tz_timer.tick()
    while tz_timer.tick() - start_time < how_long :
        if  not me.io :
            me.reopen()
            try     :
                me.start_usb()
            except ( a_cms50_exception, a_cms50_timeout_exception, ) :
                pass
            pass
        try         :
            s       = me.read_sample(progress_rtn = prg.show_progress, verbose = verbose)
            if  s   :
                yy  = s.y

                if  False :
                    msa.append(s.y)
                    mss        += s.y
                    mss        -= msa[0]
                    del(msa[0])
                else            :
                    mss         = 64
                if  False       :
                    yy          = ((s.y) * ((s.bc & 0xf) + 1)) / 4.5        # tends to flatten out in the middle when there is a change to low amplitude waves
                    if  False :
                        if  yy  < 0 :
                            yy  = -math.log(-yy + math.e)
                        else    :
                            yy  =  math.log( yy + math.e)
                        pass
                    pass

                mx  = max(mx, yy)
                mn  = min(mn, yy)

                if  s.ac & 3 :
                    # print("@@@@ ac=%u" % s.ac)
                    lay.append(s.ac)

                if  not  finger :
                    samples.append(a_finger_sample(True))
                samples.append(s)
                if  ofile_name  :
                    if  fo      :
                        fo.write("%s\n" % s.csv_str())
                    samples.write_new_samples()
                ys      = (' ' * int((100.0 * (yy - mn)) / max(1.0, (mx - mn)))) + '*'
                # avya  = [ ox for ox in samples.samples[-5 * SAMPLE_RATE : ] if hasattr(ox, 'y') ]
                # avy   = sum([ ox.y for ox in avya ]) / float(max(1, len(avya)))

                print("%s %s" % ( s.print_str(), ys ))              # (100.0 * yy) / (yy + (s.bc & 0xf)), ys )

                if  not finger :
                    finger  = True
                    samples.flush_file()
                    print("Finger")
                stopped = max(stopped - 10, 0)

                if  len(samples.samples) > png_drs[-1] * SAMPLE_RATE * 2 :
                    samples.flush_file()
                    samples.forget_old_samples(png_drs[-1] * SAMPLE_RATE)
                rx_when = tzlib.elapsed_time()
            t   = tzlib.elapsed_time()
            if  t - ts > 59 :
                ts  = t
                samples.flush_file()
            pass
        except   a_cms50_no_finger_exception :
            if  finger  :
                finger  = False
                samples.append(a_finger_sample(False))
                print("No finger")
                samples.flush_file()
            pass
        except   a_cms50_data_exception as msg :
            samples.flush_file()
            print(msg)                                  # those bits are not, apparently, dupes of each other
            if  not stopped :
                sys.exit(199)
            stopped    -= 1
        except ( a_cms50_exception, a_cms50_timeout_exception, ) :
            t   = tzlib.elapsed_time()
            if  t - rx_when > 0.5 :
                rx_when = t
                ts      = t
                mx      = -10000
                mn      =  10000
                samples.flush_file()
                # print("@@@@ main::exception")
                tzlib.print_exception()
                try     :
                    me.start_usb()
                except a_cms50_timeout_exception :
                    # print("@@@@ main start_usb timeout exception")
                    pass
                except a_cms50_exception :      # as e :
                    me.blind_close()
                    # print("@@@@", "\n".join(dir(e.args)))
                    # print("@@@@ main start_usb '%s' exception" % e.args)
                    pass
                pass
            pass

        if  len(me.data) :
            samples.flush_file()
            while len(me.data) :
                dt  = me.data.pop(0)
                # print("@@@@", len(dt.samples), len(dt.fd), "at", dt.tm // 3600, (dt.tm // 60) % 60)
                if  len(dt.samples) :
                    fn  = get_output_file_name(ofile_name, program_name = program_name, ext = ".dat")
                    # tzlib.write_whole_binary_file(fn, dt.fd)
                    dfn = dt.write_file(fn)
                    if  not dfn :
                        print("Probably no data to write, so file not written.")
                    else :
                        print("Wrote driven upload to", fn, "and", dfn, len(dt.samples))
                    pass
                pass
            mx      = -10000
            mn      =  10000
            ts      = tzlib.elapsed_time()
            stopped = 1000

        k   = TZKeyReady.key_ready()
        if  k :
            print('')

            if  k in [ 'q', ] :
                break

            if  k == '?' :
                pcmd    = ("""
p       Write %s file with latest data.
""" % ( a_recording.FILE_EXT, )).rstrip()
                print("""
q       Quit.
ESC     Quit.
?       Help.
r       Reset some values.
+       Up    the .png duration - currently %s
-       Lower the .png duration.%s
d       Wait for data upload to write to %s file.

""" % ( hms(png_drs[png_di]), pcmd, a_recording.FILE_EXT )
                     )
                print("ymx=%d ymn=%d lay=%s %.1f samples per second" % ( mx, mn, str(lay), me.scnt / max(1, (tzlib.elapsed_time() - ts)) ))      # 60 per second

            if  k == 'r' :
                samples.flush_file()
                mx      = -10000
                mn      =  10000
                ts      = tzlib.elapsed_time()
                lay     = []
                me.scnt = 1

            if  k == 'd' :
                samples.flush_file()
                stopped = 1000
                fn      = get_output_file_name(ofile_name, program_name = program_name, ext = ".dat")
                print("Uploading to", fn, "You may need to turn on 'upload' in the menus.")
                dt      = me.read_upload(progress_rtn = prg.show_progress, verbose = verbose)
                # print("@@@@", len(fd), "at", tm // 3600, (tm // 60) % 60)
                if  dt and len(dt.samples) :
                    # tzlib.write_whole_binary_file(fn, dt.fd)
                    dfn = dt.write_file(fn)
                    if  not dfn :
                        print("Probably no data to write, so file not written.")
                    else :
                        print("Wrote upload to", fn, "and", dfn, len(dt.samples))
                    pass
                else    :
                    print("Failed to upload.")
                pass


            if  k == 'E' :
                print("Sending E")
                me.tx("E")
            if  k == 'F' :
                print("Sending F")
                me.tx("F")
            if  k == 'G' :
                print("Sending G")
                me.tx("G")
            if  k == 'H' :
                print("Sending H")
                me.tx("H")

            if  k == 's' :
                if  False :
                    me.tx(chr(sb))
                    print("s %02x:%u" % ( sb, sb ))
                    sb -= 1
                    if  sb  < 0 :
                        sb  = 255
                    pass
                else :
                    me.tx("\xf5\xf5")
                pass

            if  k == 'S' :
                if  False :
                    me.tx(chr(sb))
                    print("s %02x:%u" % ( sb, sb ))
                    sb += 1
                    if  sb  > 255 :
                        sb  = 0
                    pass
                else :
                    me.tx("\xf6\xf6\xf6")
                pass


            if  k in [ '+', '=' ] :
                png_di  = min(len(png_drs) - 1, png_di + 1)
                print(".png duration %s" % hms(png_drs[png_di]))
            if  k in [ '-', '_' ] :
                png_di  = max(0, png_di - 1)
                print(".png duration %s" % hms(png_drs[png_di]))
            if  k == 'p' :
                if  (len(samples.samples) > 60) and (mn < mx) :
                    fn  = get_output_file_name(ofile_name, program_name = program_name, ext = ".png")

                    sma = [ sa for sa in samples.samples if hasattr(sa, 'ox') ]
                    tw  = min(int(png_drs[png_di] * SAMPLE_RATE), len(sma))
                    sma = sma[len(sma) - tw : ]

                    if  not len(sma) :
                        "No data to write/graph"
                    else :
                        dt  = a_recording(samples = sma)
                        dfn = dt.write_file(fn)
                        if  not dfn :
                            print("Probably no data to write, so file not written.")
                        else :
                            print("Wrote %s" % dfn)
                        pass
                    pass
                pass

            pass
        pass

    if  ofile_name  :
        if  fo      :
            fo.close()
        samples.close_write_file()

    me.blind_close()


if  __name__ == '__main__' :

    main("")


#
#
# eof
